/*----------------------------------------------------------------------------
                        autoscale.cc (autoscaling axes)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "frame.h"
#include "axis.h"
#include "naninf.h"
#include "topazvalues.h"

extern void message(const char *msg);
extern int CalcFormulastring(char *s, double *n, int InitFormula, data* dat, int cellno, int seriesno, int Isminmax);

void frame::autoscaleunit(axis *AUnit, double Start, double End, bool Spacing, int Spacingfactor)
{
  double SpaceMargin;
  double dx;
  double DMin, DMax, DUnit;
  double ldx, fnx;
  
  if (fabs(Start) >= MAXDOUBLE || fabs(End) >= MAXDOUBLE)
    return;
  
  if (Start == End)
    return;
  
  if (Spacingfactor < 0 || Spacingfactor >= 50)
    return;

  if (Spacing)
    SpaceMargin = (double)Spacingfactor / 100.0;
  else
    SpaceMargin = 0;
  
  switch(AUnit->scaling)
    {
      //Linear
    case SCALE_LINEAR:
      DMin = Start;
      DMax = End;
      dx = End - Start;
      ldx = floor(log10(fabs(dx)));
      fnx = fabs(dx) / pow(10.0,ldx);
      DMin -= fabs(dx) * SpaceMargin;
      DMax += fabs(dx) * SpaceMargin;
      if (fnx <= 2.0)
	DUnit = pow(10.0, ldx) / 2.0;
      else if (fnx > 2.0 && fnx <= 5.0)
	DUnit = pow(10.0, ldx);
      else if (fnx > 5.0 && fnx <= 10.0)
	DUnit = pow(10.0, ldx) * 2.0;
      else
	DUnit = pow(10.0,ldx);
      
      AUnit->start = DMin;
      AUnit->end   = DMax;
      AUnit->unit  = DUnit;
      
      break;
      //Log
    case SCALE_LOG:
      if (Start > 0.0)
	{
	  DMin = Start;
	  DMax = End;
	  dx = log10(End) - log10(Start);
	  DMin /= pow(10.0, dx * SpaceMargin);
	  DMax *= pow(10.0, dx * SpaceMargin);
	  
	  AUnit->start          = DMin;
	  AUnit->end            = DMax;
	  AUnit->unit           = 10.0;
	}
      else
	return;
      break;
      //Inverse
    case SCALE_INV:
      if (Start > 0.0)
	{
	  DMin = Start;
	  DMax = End;
	  dx = 1.0 / Start - 1.0 / End;
	  
	  DMin = 1.0 / DMin + dx * SpaceMargin;
	  DMin = 1.0 / DMin;
	  if ((1.0 / DMin - dx * SpaceMargin) > 0.0
	      && 1.0 / (1.0 / DMax - dx * SpaceMargin) < 1e300)
	    {
	      DMax = 1.0 / DMax - dx * SpaceMargin;
	      DMax = 1.0 / DMax;
	    }
	  
	  AUnit->start    	= DMin;
	  AUnit->end    	= DMax;
	  AUnit->unit   	= 10.0;
	}
      else
	return;
      break;
      //log10(Linear)
    case SCALE_LOGLINEAR:
      if (Start > 0.0)
	{
	  DMin = Start;
	  DMax = End;
	  dx = log10(End) - log10(Start);
	  DMin /= pow(10.0, dx * SpaceMargin);
	  DMax *= pow(10.0, dx * SpaceMargin);
	  
	  dx = DMax - DMin;
	  ldx = floor(log10(fabs(dx)));
	  fnx = fabs(dx) / pow(10.0,ldx);
	  
	  if (fnx <= 2.0)
	    DUnit = pow(10.0, ldx) / 2.0;
	  else if (fnx > 2.0 && fnx <= 5.0)
	    DUnit = pow(10.0, ldx);
	  else if (fnx > 5.0 && fnx <= 10.0)
	    DUnit = pow(10.0, ldx) * 2.0;
	  else
	    DUnit = pow(10.0,ldx);
	  
	  AUnit->start  	= DMin;
	  AUnit->end    	= DMax;
	  AUnit->unit   	= DUnit;
	}
      else
	return;
      break;
      //inverse(Linear)
    case SCALE_INVLINEAR:
      if (Start > 0.0)
	{
	  DMin = Start;
	  DMax = End;
	  dx = 1.0 / Start - 1.0 / End;
	  
	  DMin = 1.0 / DMin + dx * SpaceMargin;
	  DMin = 1.0 / DMin;
	  if ((1.0 / DMin - dx * SpaceMargin) > 0.0
	      && 1.0 / (1.0 / DMax - dx * SpaceMargin) < 1e300)
	    {
	      DMax = 1.0 / DMax - dx * SpaceMargin;
		    DMax = 1.0 / DMax;
	    }
	  
	  dx = DMax - DMin;
	  ldx = floor(log10(fabs(dx)));
		fnx = fabs(dx) / pow(10.0,ldx);
	  
	  if (fnx <= 2.0)
	    DUnit = pow(10.0, ldx) / 2.0;
		else if (fnx > 2.0 && fnx <= 5.0)
		  DUnit = pow(10.0, ldx);
		else if (fnx > 5.0 && fnx <= 10.0)
		  DUnit = pow(10.0, ldx) * 2.0;
		else
		  DUnit = pow(10.0,ldx);
	  
	  AUnit->start  	= DMin;
	  AUnit->end    	= DMax;
	  AUnit->unit   	= DUnit;
	}
      break;
      
    }
}




void frame::autoscale(bool BFlag, bool LFlag, bool TFlag, bool RFlag, bool Spacing, int Spacingfactor)
{
  data *PPlot = 0;
  int plotno = darray->getarraylength();
  int i, j;
  std::string temp_pre_formula;
  double fa;
  int NumPoint;
  double MinX, MaxX, MinY, MaxY, PosiMinX, PosiMaxX, PosiMinY, PosiMaxY;
  double MinT, MaxT, MinR, MaxR, PosiMinT, PosiMaxT, PosiMinR, PosiMaxR;
  
  
  MinX 		= MAXDOUBLE;
  MaxX 		= -MAXDOUBLE;
  MinY 		= MAXDOUBLE;
  MaxY 		= -MAXDOUBLE;
  PosiMinX 	= MAXDOUBLE;
  PosiMaxX 	= -MAXDOUBLE;
  PosiMinY 	= MAXDOUBLE;
  PosiMaxY 	= -MAXDOUBLE;
  
  MinT 		= MAXDOUBLE;
  MaxT 		= -MAXDOUBLE;
  MinR 		= MAXDOUBLE;
  MaxR 		= -MAXDOUBLE;
  PosiMinT 	= MAXDOUBLE;
  PosiMaxT 	= -MAXDOUBLE;
  PosiMinR 	= MAXDOUBLE;
  PosiMaxR 	= -MAXDOUBLE;
  
  for (i = 0; i < plotno; i++)
    {
      if ((*darray)[i].display)
	{
	  
	  PPlot = &(*darray)[i];

	  //input to buffer
	  NumPoint = PPlot->x->getarraylength() - 1;
	  
	  //Xdata
	  if (PPlot->trans_x)
	    temp_pre_formula = *PPlot->format_x;
	  else
	    temp_pre_formula = std::string("x");
	  
	  CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, true ,PPlot, 0, getdataindex(PPlot->id), 0);
	  for (j = 0; j <= NumPoint; j++)
	    {
	      if ((*PPlot->x)[j].status == 0 && (*PPlot->y)[j].status == 0)
                {
		  CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, false ,PPlot, j, getdataindex(PPlot->id), 0);
		  if (isnormalvalue(fa))
		    {
		      if (getaxisindex(PPlot->refaxis_x) == 0)
			{
			  if (fa <= MinX)
			    MinX = fa;
			  if (fa >= MaxX)
			    MaxX = fa;
			  if (fa > 0.0 && fa <= PosiMinX)
			    PosiMinX = fa;
			  if (fa > 0.0 && fa >= PosiMaxX)
			    PosiMaxX = fa;
			}
		      else if (getaxisindex(PPlot->refaxis_x) == 2)
			{
			  if (fa <= MinT)
			    MinT = fa;
			  if (fa >= MaxT)
			    MaxT = fa;
			  if (fa > 0.0 && fa <= PosiMinT)
			    PosiMinT = fa;
			  if (fa > 0.0 && fa >= PosiMaxT)
			    PosiMaxT = fa;
			}
		    }
            	}
	    }
	  
	  //Ydata
	  if (PPlot->trans_y)
	    temp_pre_formula = *PPlot->format_y;
	  else
	    temp_pre_formula = std::string("y");
	  
	  CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, true ,PPlot, 0, getdataindex(PPlot->id), 0);
	  for (j = 0; j <= NumPoint; j++)
	    {
	      if ((*PPlot->x)[j].status == 0 && (*PPlot->y)[j].status == 0)
                {
		  CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, false ,PPlot, j, getdataindex(PPlot->id), 0);
		  if (isnormalvalue(fa))
		    {
		      if (getaxisindex(PPlot->refaxis_y) == 1)
			{
			  if (fa <= MinY)
			    MinY = fa;
			  if (fa >= MaxY)
			    MaxY = fa;
			  if (fa > 0.0 && fa <= PosiMinY)
			    PosiMinY = fa;
			  if (fa > 0.0 && fa >= PosiMaxY)
			    PosiMaxY = fa;
			}
		      else if (getaxisindex(PPlot->refaxis_y) == 3)
			{
			  if (fa <= MinR)
			    MinR = fa;
			  if (fa >= MaxR)
			    MaxR = fa;
			  if (fa > 0.0 && fa <= PosiMinR)
			    PosiMinR = fa;
			  if (fa > 0.0 && fa >= PosiMaxR)
			    PosiMaxR = fa;
			}
		    }
            	}
	    }
	  if (BFlag)
	    {
	      switch((*axarray)[0].scaling)
		{
		case 0:
		  autoscaleunit(&(*axarray)[0], MinX, MaxX, Spacing, Spacingfactor);
		  break;
		case 1:
		case 2:
		case 3:
		case 4:
		  if (PosiMaxX > 0)
		    autoscaleunit(&(*axarray)[0], PosiMinX, PosiMaxX, Spacing, Spacingfactor);
		  break;
		}
	    }
	  if (LFlag)
	    {
	      switch((*axarray)[1].scaling)
		{
		case 0:
		  autoscaleunit(&(*axarray)[1], MinY, MaxY, Spacing, Spacingfactor);
		  break;
		case 1:
		case 2:
		case 3:
		case 4:
		  if (PosiMaxY > 0)
		    autoscaleunit(&(*axarray)[1], PosiMinY, PosiMaxY, Spacing, Spacingfactor);
		  break;
		}
	    }
	  if (TFlag)
	    {
	      switch((*axarray)[2].scaling)
		{
		case 0:
		  autoscaleunit(&(*axarray)[2], MinT, MaxT, Spacing, Spacingfactor);
		  break;
		case 1:
		case 2:
		case 3:
		case 4:
		  if (PosiMaxT > 0)
		    autoscaleunit(&(*axarray)[2], PosiMinT, PosiMaxT, Spacing, Spacingfactor);
		  break;
		}
	    }
	  if (RFlag)
	    {
	      switch((*axarray)[3].scaling)
		{
		case 0:
		  autoscaleunit(&(*axarray)[3], MinR, MaxR, Spacing, Spacingfactor);
		  break;
		case 1:
		case 2:
		case 3:
		case 4:
		  if (PosiMaxR > 0)
		    autoscaleunit(&(*axarray)[3], PosiMinR, PosiMaxR, Spacing, Spacingfactor);
		  break;
		}
	    }
	  arrangeaxes();
	}

    }
}

