/*----------------------------------------------------------------------------
                           axis.cc (axis object)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "axis.h"
#include "frame.h"
#include "graph.h"
#include "col.h"
#include "gdi.h"
#include "minmax.h"
#include "pointdefs.h"
#include "naninf.h"
#include "filepoint.h"
#include "topazvalues.h"

extern void message(const char *msg);
extern void StyleLinePath(bool flag, FILE* f, doublepointarray *array, int no_of_point, int style, int width,
			  bool Clip, intrect *AxisRect, intrect *boundrect);

extern FileHandleArray *fhvuff;


double rot_x(double x, double y, double th)
{
  double ra = th / 180.0 * M_PI;

  return (x * cos(ra) + y * sin(ra));
}
double rot_y(double x, double y, double th)
{
  double ra = th / 180.0 * M_PI;

  return (-x * sin(ra) + y * cos(ra));
}

void BBboxrotatedrect(intrect *r, double th)
{
  // origin of rotation is assumed to be (0, 0)
  
  double minx, maxx, miny, maxy;
  double x, y;

  minx = rot_x((double)r->left, (double)r->top, th);
  maxx = rot_x((double)r->left, (double)r->top, th);
  miny = rot_y((double)r->left, (double)r->top, th);
  maxy = rot_y((double)r->left, (double)r->top, th);
  
  x = rot_x((double)r->right, (double)r->top, th);
  y = rot_y((double)r->right, (double)r->top, th);

  minx = min(minx, x);
  maxx = max(maxx, x);
  miny = min(miny, y);
  maxy = max(maxy, y);

  x = rot_x((double)r->right, (double)r->bottom, th);
  y = rot_y((double)r->right, (double)r->bottom, th);

  minx = min(minx, x);
  maxx = max(maxx, x);
  miny = min(miny, y);
  maxy = max(maxy, y);

  x = rot_x((double)r->left, (double)r->bottom, th);
  y = rot_y((double)r->left, (double)r->bottom, th);

  minx = min(minx, x);
  maxx = max(maxx, x);
  miny = min(miny, y);
  maxy = max(maxy, y);

  r->left   = (int)minx;
  r->top    = (int)miny;
  r->right  = (int)maxx;
  r->bottom = (int)maxy;

}

axis::axis():plobj()
{
  version              = 4;
  id                   = 0;
  label                = new std::string("");
  selected             = false;

  axistype             = 0;
  display              = true;
  commonset            = true;
  scaling              = SCALE_LINEAR;
  invertscale          = false;
  origin.x             = 0;
  origin.y             = 0;
  length               = 10000;
  angle                = 0.0;
  baselinestyle.init();
  basecolor.setblack();
  tickarea_start       = 0.0;
  tickarea_end         = 100.0;

  /*mumerical range*/
  start                = 0.0;
  end                  = 10.0;
  unit                 = 2.0;
  crosspoint           = 0.0;
  autocrosspoint       = true;

  /*tick*/
  tickparity           = TICKPARITY_PLUS;
  tickwidth            = 10;
  ticklength           = 200;           
  division             = 2;

  /*tickline*/
  majorticklinecolor.setblack();
  minorticklinecolor.setblack();
  majorticklinestyle.init();
  majorticklinestyle.style = LINESTYLENULL;
  minorticklinestyle.init();
  minorticklinestyle.style = LINESTYLENULL;

  /*label*/
  fo.init();
  fontcolor.setblack();
  format                = LABEL_FORMAT_DECIMAL;
  addplus               = false;
  offset                = 0;
  skip                  = 0;
  autoprecision         = true;
  intprecision          = 5;
  decimalprecision      = 5;
  labelparity           = LABELPARITY_PLUS;
  crosspointparity      = CROSSPOINTPARITY_CENTER;
  crosspointoffset      = 0;
}

axis::~axis()
{
  delete label;
}

axis& axis::operator = (const axis&  axis)
{
//  id                   = axis.id;
  selected             = axis.selected;
  axistype             = axis.axistype;
  display              = axis.display;
  commonset            = axis.commonset;
  scaling              = axis.scaling;
  invertscale          = axis.invertscale;
  origin.x             = axis.origin.x;
  origin.y             = axis.origin.y;
  length               = axis.length;
  angle                = axis.angle;
  baselinestyle        = axis.baselinestyle;
  basecolor            = axis.basecolor;
  tickarea_start       = axis.tickarea_start;
  tickarea_end         = axis.tickarea_end;


  /*mumerical range*/
  start                = axis.start;
  end                  = axis.end;
  unit                 = axis.unit;
  crosspoint           = axis.crosspoint;
  autocrosspoint       = axis.autocrosspoint;

  /*tick*/
  tickparity           = axis.tickparity;
  tickwidth            = axis.tickwidth;
  ticklength           = axis.ticklength;           
  division             = axis.division;

  /*tickline*/
  majorticklinecolor   = axis.majorticklinecolor;
  minorticklinecolor   = axis.minorticklinecolor;
  majorticklinestyle   = axis.majorticklinestyle;
  minorticklinestyle   = axis.minorticklinestyle;

  /*label*/
  fo                   = axis.fo;
  fontcolor            = axis.fontcolor;
  format               = axis.format;
  addplus              = axis.addplus;
  offset               = axis.offset;
  skip                 = axis.skip;
  autoprecision        = axis.autoprecision;
  intprecision         = axis.intprecision;
  decimalprecision     = axis.decimalprecision;
  labelparity          = axis.labelparity;
  crosspointparity     = axis.crosspointparity;
  crosspointoffset     = axis.crosspointoffset;

  return *this;
}


bool axis::get(std::string* member, tokenbuff* outvalue)
{
  std::string out;
  int iid;
  char stemp[100];

  if (topobject(member, &out, &iid) == true)
    {
      if (out == std::string("id") && iid == -1)
	{
	  sprintf(stemp, "%d", id);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("label") && iid == -1)
	{
	  outvalue->value->add(label->c_str());
	  return true;
	}
      else if (out == std::string("selected") && iid == -1)
	{
	  sprintf(stemp, "%d", selected);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("axistype") && iid == -1)
	{
	  sprintf(stemp, "%d", axistype);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("show") && iid == -1)
	{
	  sprintf(stemp, "%d", display);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("commonset") && iid == -1)
	{
	  sprintf(stemp, "%d", commonset);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("scaling") && iid == -1)
	{
	  sprintf(stemp, "%d", scaling);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("invertscale") && iid == -1)
	{
	  sprintf(stemp, "%d", invertscale);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("x") && iid == -1)
	{
	  sprintf(stemp, "%d", origin.x);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("y") && iid == -1)
	{
	  sprintf(stemp, "%d", origin.y);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("length") && iid == -1)
	{
	  sprintf(stemp, "%d", length);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("angle") && iid == -1)
	{
	  sprintf(stemp, "%.15g", angle);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("baselinewidth") && iid == -1)
	{
	  sprintf(stemp, "%d", baselinestyle.width);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("baselinecap") && iid == -1)
	{
	  sprintf(stemp, "%d", baselinestyle.cap);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("baselinejoin") && iid == -1)
	{
	  sprintf(stemp, "%d", baselinestyle.join);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("baselinestyle") && iid == -1)
	{
	  sprintf(stemp, "%d", baselinestyle.style);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("baselinecolor_red") && iid == -1)
	{
	  sprintf(stemp, "%u", basecolor.red);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("baselinecolor_green") && iid == -1)
	{
	  sprintf(stemp, "%u", basecolor.green);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("baselinecolor_blue") && iid == -1)
	{
	  sprintf(stemp, "%u", basecolor.blue);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("tickarea_start") && iid == -1)
	{
	  sprintf(stemp, "%.15g", tickarea_start);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("tickarea_end") && iid == -1)
	{
	  sprintf(stemp, "%.15g", tickarea_end);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("start") && iid == -1)
	{
	  sprintf(stemp, "%.15g", start);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("end") && iid == -1)
	{
	  sprintf(stemp, "%.15g", end);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("unit") && iid == -1)
	{
	  sprintf(stemp, "%.15g", unit);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("crosspoint") && iid == -1)
	{
	  sprintf(stemp, "%.15g", crosspoint);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("autocrosspoint") && iid == -1)
	{
	  sprintf(stemp, "%d", autocrosspoint);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("tickparity") && iid == -1)
	{
	  sprintf(stemp, "%d", tickparity);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("tickwidth") && iid == -1)
	{
	  sprintf(stemp, "%d", tickwidth);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("ticklength") && iid == -1)
	{
	  sprintf(stemp, "%d", ticklength);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("tickdivision") && iid == -1)
	{
	  sprintf(stemp, "%d", division);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("majorticklinecolor_red") && iid == -1)
	{
	  sprintf(stemp, "%u", majorticklinecolor.red);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("majorticklinecolor_green") && iid == -1)
	{
	  sprintf(stemp, "%u", majorticklinecolor.green);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("majorticklinecolor_blue") && iid == -1)
	{
	  sprintf(stemp, "%u", majorticklinecolor.blue);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("minorticklinecolor_red") && iid == -1)
	{
	  sprintf(stemp, "%u", minorticklinecolor.red);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("minorticklinecolor_green") && iid == -1)
	{
	  sprintf(stemp, "%u", minorticklinecolor.green);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("minorticklinecolor_blue") && iid == -1)
	{
	  sprintf(stemp, "%u", minorticklinecolor.blue);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("majorticklinewidth") && iid == -1)
	{
	  sprintf(stemp, "%d", majorticklinestyle.width);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("majorticklinecap") && iid == -1)
	{
	  sprintf(stemp, "%d", majorticklinestyle.cap);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("majorticklinejoin") && iid == -1)
	{
	  sprintf(stemp, "%d", majorticklinestyle.join);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("majorticklinestyle") && iid == -1)
	{
	  sprintf(stemp, "%d", majorticklinestyle.style);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("minorticklinewidth") && iid == -1)
	{
	  sprintf(stemp, "%d", minorticklinestyle.width);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("minorticklinecap") && iid == -1)
	{
	  sprintf(stemp, "%d", minorticklinestyle.cap);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("minorticklinejoin") && iid == -1)
	{
	  sprintf(stemp, "%d", minorticklinestyle.join);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("minorticklinestyle") && iid == -1)
	{
	  sprintf(stemp, "%d", minorticklinestyle.style);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("fontface") && iid == -1)
	{
	  sprintf(stemp, "%d", fo.face);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("fontsize") && iid == -1)
	{
	  sprintf(stemp, "%d", fo.size);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("fontangle") && iid == -1)
	{
	  sprintf(stemp, "%.15g", fo.angle);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("fontcolor_red") && iid == -1)
	{
	  sprintf(stemp, "%u", fontcolor.red);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("fontcolor_green") && iid == -1)
	{
	  sprintf(stemp, "%u", fontcolor.green);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("fontcolor_blue") && iid == -1)
	{
	  sprintf(stemp, "%u", fontcolor.blue);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("format") && iid == -1)
	{
	  sprintf(stemp, "%d", format);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("addplus") && iid == -1)
	{
	  sprintf(stemp, "%d", addplus);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("offset") && iid == -1)
	{
	  sprintf(stemp, "%d", offset);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("skip") && iid == -1)
	{
	  sprintf(stemp, "%d", skip);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("autoprecision") && iid == -1)
	{
	  sprintf(stemp, "%d", autoprecision);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("intprecision") && iid == -1)
	{
	  sprintf(stemp, "%d", intprecision);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("decimalprecision") && iid == -1)
	{
	  sprintf(stemp, "%d", decimalprecision);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("labelparity") && iid == -1)
	{
	  sprintf(stemp, "%d", labelparity);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("crosspointparity") && iid == -1)
	{
	  sprintf(stemp, "%d", crosspointparity);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("crosspointoffset") && iid == -1)
	{
	  sprintf(stemp, "%d", crosspointoffset);
	  outvalue->value->add(stemp);
	  return true;
	}
      else
	return false;
    }
  else
    return false;
      
  return true;

}


bool axis::set(std::string* member, std::string* setvalue)
{
  std::string out;
  int iid;

  if (topobject(member, &out, &iid) == true)
    {
      if (out == std::string("id") && iid == -1)
	{
	  id = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("label") && iid == -1)
	{
	  *label = *setvalue;
	  return true;
	}
      else if (out == std::string("selected") && iid == -1)
	{
	  selected = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("axistype") && iid == -1)
	{
	  axistype = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("show") && iid == -1)
	{
	  display = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("commonset") && iid == -1)
	{
	  commonset = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("scaling") && iid == -1)
	{
	  scaling = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("invertscale") && iid == -1)
	{
	  invertscale = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("x") && iid == -1)
	{
	  origin.x = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("y") && iid == -1)
	{
	  origin.y = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("length") && iid == -1)
	{
	  length = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("angle") && iid == -1)
	{
	  angle = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("baselinewidth") && iid == -1)
	{
	  baselinestyle.width = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("baselinecap") && iid == -1)
	{
	  baselinestyle.cap = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("baselinejoin") && iid == -1)
	{
	  baselinestyle.join = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("baselinestyle") && iid == -1)
	{
	  baselinestyle.style = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("baselinecolor_red") && iid == -1)
	{
	  basecolor.red = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("baselinecolor_green") && iid == -1)
	{
	  basecolor.green = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("baselinecolor_blue") && iid == -1)
	{
	  basecolor.blue = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("tickarea_start") && iid == -1)
	{
	  tickarea_start = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("tickarea_end") && iid == -1)
	{
	  tickarea_end = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("start") && iid == -1)
	{
	  start = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("end") && iid == -1)
	{
	  end = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("unit") && iid == -1)
	{
	  unit = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("crosspoint") && iid == -1)
	{
	  crosspoint = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("autocrosspoint") && iid == -1)
	{
	  autocrosspoint = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("tickparity") && iid == -1)
	{
	  tickparity = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("tickwidth") && iid == -1)
	{
	  tickwidth = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("ticklength") && iid == -1)
	{
	  ticklength = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("tickdivision") && iid == -1)
	{
	  division = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("majorticklinecolor_red") && iid == -1)
	{
	  majorticklinecolor.red = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("majorticklinecolor_green") && iid == -1)
	{
	  majorticklinecolor.green = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("majorticklinecolor_blue") && iid == -1)
	{
	  majorticklinecolor.blue = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("minorticklinecolor_red") && iid == -1)
	{
	  minorticklinecolor.red = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("minorticklinecolor_green") && iid == -1)
	{
	  minorticklinecolor.green = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("minorticklinecolor_blue") && iid == -1)
	{
	  minorticklinecolor.blue = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("majorticklinewidth") && iid == -1)
	{
	  majorticklinestyle.width = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("majorticklinecap") && iid == -1)
	{
	  majorticklinestyle.cap = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("majorticklinejoin") && iid == -1)
	{
	  majorticklinestyle.join = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("majorticklinestyle") && iid == -1)
	{
	  majorticklinestyle.style = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("minorticklinewidth") && iid == -1)
	{
	  minorticklinestyle.width = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("minorticklinecap") && iid == -1)
	{
	  minorticklinestyle.cap = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("minorticklinejoin") && iid == -1)
	{
	  minorticklinestyle.join = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("minorticklinestyle") && iid == -1)
	{
	  minorticklinestyle.style = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("fontface") && iid == -1)
	{
	  fo.face = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("fontsize") && iid == -1)
	{
	  fo.size = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("fontangle") && iid == -1)
	{
	  fo.angle = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("fontcolor_red") && iid == -1)
	{
	  fontcolor.red = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("fontcolor_green") && iid == -1)
	{
	  fontcolor.green = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("fontcolor_blue") && iid == -1)
	{
	  fontcolor.blue = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("format") && iid == -1)
	{
	  format = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("addplus") && iid == -1)
	{
	  addplus = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("offset") && iid == -1)
	{
	  offset = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("skip") && iid == -1)
	{
	  skip = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("autoprecision") && iid == -1)
	{
	  autoprecision = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("intprecision") && iid == -1)
	{
	  intprecision = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("decimalprecision") && iid == -1)
	{
	  decimalprecision = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("labelparity") && iid == -1)
	{
	  labelparity = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("crosspointparity") && iid == -1)
	{
	  crosspointparity = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("crosspointoffset") && iid == -1)
	{
	  crosspointoffset = atoi(setvalue->c_str());
	  return true;
	}
      else
	return false;
    }
  else
    return false;
      
  return true;

}


bool axis::exec(std::string* function, buffarray *argument, tokenbuff* outvalue)
{
  std::string out, shrunk, str1;
  int iid, arglength;
  intpoint ip;
  char stemp[1000];
  double ftemp;
  frame *fr = (frame *)parent;
  FILE* fp;
  int itemp;

  outvalue->value->setbuff(0);
  if (topobject(function, &out, &iid) == true)
    {
      arglength = argument->getarraylength();

      // paint
      if (out == std::string("paint") && iid == -1)
	{
	  if (arglength == 1)
	    {
	      if (fhvuff->get(&(*argument)[0], &fp, &itemp, &str1))
		{
		  paint(true, fp);
		  return true;
		}
	      return false;
	    }
	  else
	    return false;
	}
      // setbbox
      else if (out == std::string("setbbox") && iid == -1)
	{
	  if (arglength == 1 && (*argument)[0] == std::string("null"))
	    {
	      paint(false, fp);
	      return true;
	    }
	  else
	    return false;
	}
      // virtual to world
      else if (out == std::string("virtow") && iid == -1)
	{
	  if (arglength == 1)
	    {
	      ip = virtual_to_world(atof((*argument)[0].c_str()));
	      outvalue->value->setbuff(0);
	      sprintf(stemp, "%d", ip.x);  
	      outvalue->value->add(stemp);
	      sprintf(stemp, "%d", ip.y);  
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      // virtual to view
      else if (out == std::string("virtov") && iid == -1)
	{
	  if (arglength == 1)
	    {
	      outvalue->value->setbuff(0);
	      ip = virtual_to_world(atof((*argument)[0].c_str()));
	      sprintf(stemp, "%d", fr->wtov_x(ip.x));  
	      outvalue->value->add(stemp);
	      sprintf(stemp, "%d", fr->wtov_y(ip.y));  
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      // world to virtual
      else if (out == std::string("wtovir") && iid == -1)
	{
	  if (arglength == 2)
	    {
	      ftemp = doubleworld_to_virtual(atof((*argument)[0].c_str()), atof((*argument)[1].c_str()));
	      sprintf(stemp, "%.15g", ftemp);  
	      outvalue->value->setbuff(0);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      // view to virtual
      else if (out == std::string("vtovir") && iid == -1)
	{
	  if (arglength == 2)
	    {
	      ftemp = doubleworld_to_virtual(fr->vtow_x(atoi((*argument)[0].c_str())),
					     fr->vtow_y(atoi((*argument)[1].c_str())));
	      sprintf(stemp, "%.15g", ftemp);  
	      outvalue->value->setbuff(0);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else
	return false;
    }
  else
    return false;
      
  return true;
}


bool axis::paint(bool flag, FILE* f)
{
  frame *fra = (frame *)parent;
  
  if (display == false)
    return true;

  plinestyle(flag, f, fra->scale(baselinestyle.width), baselinestyle.cap, baselinestyle.join, baselinestyle.style);
  psetforecolor(flag, f, basecolor.red, basecolor.green, basecolor.blue);

  //baseline
  switch(fra->frametype)
    {
    case FRAME_BOX:
    case FRAME_XY:
    case FRAME_INDIV:

      //baseline
      pbeginpath(flag, f);
      pmoveto(flag, f, fra->wtov_x(0 + origin.x), fra->wtov_y(0 + origin.y));
      plineto(flag, f, fra->wtov_x((int)rot_x(length, 0, angle) + origin.x),
	      fra->wtov_y((int)rot_y(length, 0, angle) + origin.y));
      pstroke(flag, f);
      break;
      
    }

  if (!paintmajortick(flag, f))
    return false;
  if (!paintminortick(flag, f))
    return false;
  if (!drawlabel(flag, f))
    return false;

  return true;

}

doublepoint axis::virtual_to_doubleworld(double x)
{
  double fa;

  fa = virtual_to_axis(x);
  return axis_to_world(fa, 0.0);
}

intpoint axis::virtual_to_world(double x)
{
  doublepoint fa;
  intpoint ia;

  fa = virtual_to_doubleworld(x);
  if (fabs(fa.x) >= (double)MAXSHORT)
    ia.x = MAXSHORT;
  else
    ia.x = (int)fa.x;

  if (fabs(fa.y) >= (double)MAXSHORT)
    ia.y = MAXSHORT;
  else
    ia.y = (int)fa.y;

  return ia;
}

double axis::virtual_to_axis(double x)
{
  double fa = 0.0;
  double elength = (tickarea_end - tickarea_start) / 100.0 * length;
  double estart = tickarea_start / 100.0 * length;

  switch(scaling)
    {
    case SCALE_LINEAR:
      fa =  x - start;
      fa /= end - start;
      fa = elength * fa + estart;
      break;
      
    case SCALE_LOG:
    case SCALE_LOGLINEAR:
      fa =  log10(x) - log10(start);
      fa /= log10(end) - log10(start);
      fa = elength * fa + estart;
      break;
      
    case SCALE_INV:
    case SCALE_INVLINEAR:
      fa =  1.0 / x - 1.0 / start;
      fa /= 1.0 / end - 1.0 / start;
      fa = elength * fa + estart;
      break;
    }

  if (invertscale)
      fa = length - fa;
  
  return fa;
}

doublepoint axis::axis_to_world(double x, double y)
{
  doublepoint returnpoint;

  returnpoint.x = rot_x(x, y, angle);
  returnpoint.y = rot_y(x, y, angle);

  returnpoint.x += (double)origin.x;
  returnpoint.y += (double)origin.y;

  return (returnpoint);
}

double axis::axis_to_virtual(double x)
{
  double fa = 0.0;
  double elength = (tickarea_end - tickarea_start) / 100.0 * length;
  double estart = tickarea_start / 100.0 * length;
  double eend = tickarea_end / 100.0 * length;

  if (length == 0.0)
    return infvalue();

  if (invertscale)
    {
      switch(scaling)
	{
	case SCALE_LINEAR:
	  fa =  (x - length + eend) / elength;
	  fa *= start - end;
	  fa += end;
	  break;
	  
	case SCALE_LOG:
	case SCALE_LOGLINEAR:
	  fa =  (x - length + eend)/ elength;
	  fa *= log10(start) - log10(end);
	  fa += log10(end);
	  fa = pow(10.0, fa);
	  break;
	  
	case SCALE_INV:
	case SCALE_INVLINEAR:
	  fa =  (x - length + eend)/ elength;
	  fa *= 1.0 / start - 1.0 / end;
	  fa += 1.0 / end;
	  fa = 1.0 / fa;
	  break;
	}
    }
  else
    {
      switch(scaling)
	{
	case SCALE_LINEAR:
	  fa =  (x - estart)/ elength;
	  fa *= end - start;
	  fa += start;
	  break;
	  
	case SCALE_LOG:
	case SCALE_LOGLINEAR:
	  fa =  (x - estart)/ elength;
	  fa *= log10(end) - log10(start);
	  fa += log10(start);
	  fa = pow(10.0, fa);
	  break;
	  
	case SCALE_INV:
	case SCALE_INVLINEAR:
	  fa =  (x - estart)/ elength;
	  fa *= 1.0 / end - 1.0 / start;
	  fa += 1.0 / start;
	  fa = 1.0 / fa;
	  break;
	}
    }
  return fa;
}

double axis::world_to_axis(double x, double y)
{
  doublepoint returnpoint;

  returnpoint.x = x - (double)origin.x;
  returnpoint.y = y - (double)origin.y;

  returnpoint.x = rot_x(returnpoint.x,returnpoint. y, -angle);

  return (returnpoint.x);
}

double axis::doubleworld_to_virtual(double x, double y)
{
  double fa;

  fa = world_to_axis(x, y);
  return axis_to_virtual(fa);
}

bool axis::paintmajortick(bool flag, FILE* f)
{
  frame *fra = (frame *)parent;
  double virstart, virend, startval, endval;
  double delta;
  int itercount;
  intpoint valpos, fixedpos;
  doublepoint tempdpoint;
  double fa;
  double tickstart = 0.0, tickend = 0.0;
  doublepointarray *da;
  intrect axisrect, cliprect;

  da = new doublepointarray;

  // determine tickstart, tickend
  switch(tickparity)
    {
    case TICKPARITY_NULL:
      tickstart = 0.0; tickend = 0.0;
      break;
    case TICKPARITY_PLUS:
      tickstart = 0.0; tickend = ticklength;
      break;
    case TICKPARITY_MINUS:
      tickstart = -ticklength; tickend = 0.0;
      break;
    case TICKPARITY_BOTH:
      tickstart = -ticklength; tickend = ticklength;
      break;
    }

  // tick style
  plinestyle(flag, f, fra->scale(tickwidth), baselinestyle.cap, baselinestyle.join, LINESTYLESOLID);

  //majortick
  switch(scaling)
    {
    case SCALE_LINEAR:
    case SCALE_LOGLINEAR:
    case SCALE_INVLINEAR:

      if (unit == 0.0)
	{
	  message("Invalid unit!!\n");
	  delete da;
	  return false;
	}

      virstart = floor(start / unit) * unit;
      virend = floor(end / unit) * unit;
      virend += unit;
      delta           = DELTA_AXIS * (end - start);
      startval        = start      - delta;
      endval          = end        + delta;
      itercount = (int)((virend - virstart) / unit) + 2;

      if (itercount >100)
	{
	  message("Too much ticks!!\n");
	  delete da;
	  return false;
	}

      for (int i = 0; i <= itercount; i++)
	{
	  fa = virstart + (double)i * unit;
	  if ((fa >= startval) && (fa <= endval))
	    {
	      tempdpoint = axis_to_world(virtual_to_axis(fa), tickstart);
	      valpos.x = fra->wtov_x((int)tempdpoint.x);
	      valpos.y = fra->wtov_y((int)tempdpoint.y);
	      tempdpoint = axis_to_world(virtual_to_axis(fa), tickend);
	      fixedpos.x = fra->wtov_x((int)tempdpoint.x);
	      fixedpos.y = fra->wtov_y((int)tempdpoint.y);
	      
	      pbeginpath(flag, f); 
	      da->setbuff(0);
	      da->add(new doublepoint(fixedpos.x, fixedpos.y));
	      da->add(new doublepoint(valpos.x, valpos.y));
	      StyleLinePath(flag, f, da, da->getarraylength() - 1,
			    LINESTYLESOLID, fra->scale(tickwidth),
			    false, &axisrect, &cliprect);
	      pstroke(flag, f); 
	    }
	}
      break;
    case SCALE_LOG:
    case SCALE_INV:
      virstart = pow(10.0, floor(log10(start)));
      virend   = pow(10.0, floor(log10(end)));
      virend += 1.0;
      delta           = DELTA_AXIS * (log10(end) - log10(start));
      startval        = start      / pow(10.0, delta);
      endval          = end        * pow(10.0, delta);
      
      itercount = (int)(log10(virend / virstart)) + 2;
      for (int i = 0; i <= itercount; i++)
	{
	  fa = virstart * pow(10.0, (double)i);
	  if ((fa >= startval) && (fa <= endval))
	    {
	      tempdpoint = axis_to_world(virtual_to_axis(fa), tickstart);
	      valpos.x = fra->wtov_x((int)tempdpoint.x);
	      valpos.y = fra->wtov_y((int)tempdpoint.y);
	      tempdpoint = axis_to_world(virtual_to_axis(fa), tickend);
	      fixedpos.x = fra->wtov_x((int)tempdpoint.x);
	      fixedpos.y = fra->wtov_y((int)tempdpoint.y);
	      
	      pbeginpath(flag, f); 
	      da->setbuff(0);
	      da->add(new doublepoint(fixedpos.x, fixedpos.y));
	      da->add(new doublepoint(valpos.x, valpos.y));
	      StyleLinePath(flag, f, da, da->getarraylength() - 1,
			    LINESTYLESOLID, fra->scale(tickwidth),
			    false, &axisrect, &cliprect);
	      pstroke(flag, f); 
	    }
	}
      break;
    }

  delete da;
  return true;
}

bool axis::paintminortick(bool flag, FILE* f)
{
  frame *fra = (frame *)parent;
  double virstart, virend, startval, endval;
  double delta;
  int itercount;
  intpoint valpos, fixedpos;
  doublepoint tempdpoint;
  double fa;
  double tickstart = 0.0, tickend = 0.0;
  int logminortickstart, logminortickend, logminortickunit;
  doublepointarray *da;
  intrect axisrect, cliprect;

  da = new doublepointarray;

  // determine tickstart, tickend
  switch(tickparity)
    {
    case TICKPARITY_NULL:
      tickstart = 0.0; tickend = 0.0;
      break;
    case TICKPARITY_PLUS:
      tickstart = 0.0; tickend = ticklength / 2;
      break;
    case TICKPARITY_MINUS:
      tickstart = -ticklength / 2; tickend = 0.0;
      break;
    case TICKPARITY_BOTH:
      tickstart = -ticklength / 2; tickend = ticklength / 2;
      break;
    }

  // tick style
  plinestyle(flag, f, fra->scale(tickwidth), baselinestyle.cap, baselinestyle.join, LINESTYLESOLID);

  //minortick
  switch(scaling)
    {
    case SCALE_LINEAR:
    case SCALE_LOGLINEAR:
    case SCALE_INVLINEAR:

      if (unit == 0.0)
	{
	  message("Invalid unit!!\n");
	  delete da;
	  return false;
	}

      if (division > 100)
	{
	  message("Too much division number!!\n");
	  delete da;
	  return false;
	}

      virstart = floor(start / unit) * unit;
      virend = floor(end / unit) * unit;
      virend += unit;
      delta           = DELTA_AXIS * (end - start);
      startval        = start      - delta;
      endval          = end        + delta;
      itercount = (int)((virend - virstart) / unit * (double)division) + 2;


      for (int i = 0; i <= itercount; i++)
	{
	  fa = virstart + (double)i * unit / (double)division;

	  if ((fa >= startval) && (fa <= endval))
	    {
	      if ((i % division) != 0)         // except tick position
		{
		  tempdpoint = axis_to_world(virtual_to_axis(fa), tickstart);
		  valpos.x = fra->wtov_x((int)tempdpoint.x);
		  valpos.y = fra->wtov_y((int)tempdpoint.y);
		  tempdpoint = axis_to_world(virtual_to_axis(fa), tickend);
		  fixedpos.x = fra->wtov_x((int)tempdpoint.x);
		  fixedpos.y = fra->wtov_y((int)tempdpoint.y);
		  
		  pbeginpath(flag, f); 
		  da->setbuff(0);
		  da->add(new doublepoint(fixedpos.x, fixedpos.y));
		  da->add(new doublepoint(valpos.x, valpos.y));
		  StyleLinePath(flag, f, da, da->getarraylength() - 1,
				LINESTYLESOLID, fra->scale(tickwidth),
				false, &axisrect, &cliprect);
		  pstroke(flag, f); 
		}
	    }
	}
      break;
    case SCALE_LOG:
    case SCALE_INV:
      switch(division)
	{
	case 1:
	  logminortickstart       = 1;
	  logminortickend         = -1;
	  logminortickunit        = 1;
	  break;
	case 2:
	  logminortickstart       = 5;
	  logminortickend         = 5;
	  logminortickunit        = 1;
	  break;
	case 5:
	  logminortickstart       = 2;
	  logminortickend         = 8;
	  logminortickunit        = 2;
	  break;
	case 10:
	default:
	  logminortickstart       = 2;
	  logminortickend         = 9;
	  logminortickunit        = 1;
	  break;
	}

      virstart = pow(10.0, floor(log10(start)));
      virend   = pow(10.0, floor(log10(end)));
      virend += 1.0;
      delta           = DELTA_AXIS * (log10(end) - log10(start));
      startval        = start      / pow(10.0, delta);
      endval          = end        * pow(10.0, delta);
      
      itercount = (int)(log10(virend / virstart)) + 2;

      for (int i = 0; i <= itercount; i++)
	{
	  for (int j = logminortickstart; j <= logminortickend; j = j + logminortickunit)
	    {
	      fa = virstart * pow(10.0, (double)i) * (double)j;
	      if ((fa >= startval) && (fa <= endval))
		{
		  tempdpoint = axis_to_world(virtual_to_axis(fa), tickstart);
		  valpos.x = fra->wtov_x((int)tempdpoint.x);
		  valpos.y = fra->wtov_y((int)tempdpoint.y);
		  tempdpoint = axis_to_world(virtual_to_axis(fa), tickend);
		  fixedpos.x = fra->wtov_x((int)tempdpoint.x);
		  fixedpos.y = fra->wtov_y((int)tempdpoint.y);
		  
		  pbeginpath(flag, f); 
		  da->setbuff(0);
		  da->add(new doublepoint(fixedpos.x, fixedpos.y));
		  da->add(new doublepoint(valpos.x, valpos.y));
		  StyleLinePath(flag, f, da, da->getarraylength() - 1,
				LINESTYLESOLID, fra->scale(tickwidth),
				false, &axisrect, &cliprect);
		  pstroke(flag, f); 
		}
	    }
	}
      break;
    }

  delete da;
  return true;
}


void NormalizeTickLabel(std::string *OutputText, std::string *NormFactor
			, double InitValue, double EndValue, double UnitValue, double Value
			, bool IsLinearScale, int Formula, bool IsAddPlus
			, bool AutoNumPrecision, int IntPrecision, int DecimalPrecision)
{

  bool KikakuFlag = false;
  double fkikaku = 1.0, fkikakumin, fkikakumax;
  char stemp[30];
  char stemp1[100];
  double LogValue;
  int shousuu, shisuu;
  double kasuu;


  *OutputText = std::string("");

  if(IsLinearScale)
    {
      LogValue = floor(Value / UnitValue - .001) + 1;
      
      //$B5,3J2=0x;R(B
      if (fabs(InitValue) <= fabs(EndValue))
	{
	  fkikakumax = fabs(EndValue);
	  fkikakumin = fabs(InitValue);
	}
      else
	{
	  fkikakumax = fabs(InitValue);
	  fkikakumin = fabs(EndValue);
	}
      
      sprintf(stemp,"%g", fkikakumax);
      if (strchr(stemp,'e') != 0)
	{
	  KikakuFlag = true;
	  fkikaku = fkikakumax;
	}
      else
	KikakuFlag = false;
      
      //normalize factor
      if (KikakuFlag)
	fkikaku = pow(10.0, floor(log10(fabs(fkikaku))));
      else
	fkikaku=1.0;
      
      //$B>.?tE@0J2<$N7W;;(B
      sprintf(stemp,"%.15g", UnitValue / fkikaku);
      if (strchr(stemp, '.') != 0)
	shousuu = strlen(strchr(stemp, '.')) - 1;
      else
	shousuu = 0;
      
      if (!AutoNumPrecision)
        {
	  if (DecimalPrecision == -1)
            {
	      if (IsAddPlus)
		sprintf(stemp,"%%+.%dlf",shousuu);
	      else
		sprintf(stemp,"%%.%dlf",shousuu);
            }
	  else
            {
	      if (IsAddPlus)
		sprintf(stemp,"%%+%d.%dlf",IntPrecision, DecimalPrecision);
	      else
		sprintf(stemp,"%%%d.%dlf",IntPrecision, DecimalPrecision);
            }
	  sprintf(stemp1, stemp, LogValue * UnitValue / fkikaku);
	  
        }
      else
        {
	  if (IsAddPlus)
	    sprintf(stemp1, "%+.15g", LogValue * UnitValue / fkikaku);
	  else
	    sprintf(stemp1, "%.15g", LogValue * UnitValue / fkikaku);
        }
      *OutputText = std::string(stemp1);
      
      //$B5,3J2=0x;R$NI=<((B
      if (KikakuFlag)
	{
	  sprintf(stemp1, "(#G\\264#F10^%d|)", (int)floor(log10(fkikaku) + .5));
	  *NormFactor = std::string(stemp1);
	}
      else
	{
	  *NormFactor = std::string("");
	}
    }
  else
    {
      
      LogValue = floor(log10(Value) / log10(UnitValue) - .001) + 1;

      if(Formula == 1)	//decimal
	{
	  if (IsAddPlus)
	    sprintf(stemp1, "%+.15g",pow(10.0, LogValue));
	  else
	    sprintf(stemp1, "%.15g", pow(10.0, LogValue));
	  
	}
      else if (Formula == 2)	//exp
	{
	  shisuu = (int)LogValue;
	  kasuu = (InitValue * pow(UnitValue, LogValue)) / pow(10.0, (double)shisuu);

	  if (IsAddPlus)
	    sprintf(stemp1, "10^%+d", shisuu);
	  else
	    sprintf(stemp1, "10^%d", shisuu);
	}
      *OutputText = std::string(stemp1);
    }
}


bool axis::drawlabel(bool flag, FILE* f)
{
  frame *fra = (frame *)parent;
  double virstart, virend, startval, endval;
  double delta;
  int itercount;
  intpoint valpos, fixedpos;
  doublepoint tempdpoint;
  double fa;
  double tickstart = 0.0, tickend = 0.0;
  intrect axisrect, cliprect;
  parts* t; 
  intrect ir;
  int h, w;
  std::string normfactor;
  int skipfactor, labelcountfromzero;
  int lateralposition = 0;
  double cprange = .001;
  double cp = 0.0;

  if (display == false)
    return true;

  //label is null
  if (format == LABEL_FORMAT_NULL)
    return true;
  
  t = new parts;
  t->fo = fo;
  t->fo.angle = fo.angle + angle;
  t->fontcolor = fontcolor;
  t->parent = parent;

  // set ascend
  h = (int)((double)t->getascend(&t->fo, false));

  // select textcolor
  psetforecolor(flag, f, fontcolor.red, fontcolor.green, fontcolor.blue);


  // determine tickstart, tickend
  switch(tickparity)
    {
    case TICKPARITY_NULL:
      tickstart = 0.0; tickend = 0.0;
      break;
    case TICKPARITY_PLUS:
      tickstart = 0.0; tickend = ticklength;
      break;
    case TICKPARITY_MINUS:
      tickstart = -ticklength; tickend = 0.0;
      break;
    case TICKPARITY_BOTH:
      tickstart = -ticklength; tickend = ticklength;
      break;
    }

  // tick style
  plinestyle(flag, f, fra->scale(tickwidth), baselinestyle.cap, baselinestyle.join, LINESTYLESOLID);


  normfactor = std::string("");
  //majortick
  switch(scaling)
    {
    case SCALE_LINEAR:
    case SCALE_LOGLINEAR:
    case SCALE_INVLINEAR:

      if (unit == 0.0)
	{
	  message("Invalid unit!!\n");
	  delete t;
	  return false;
	}

      virstart = floor(start / unit) * unit;
      virend = floor(end / unit) * unit;
      virend += unit;
      delta           = DELTA_AXIS * (end - start);
      startval        = start      - delta;
      endval          = end        + delta;
      itercount = (int)((virend - virstart) / unit) + 2;

      if (itercount >100)
	{
	  message("Too much ticks!!\n");
	  delete t;
	  return false;
	}

      // set skip factor
      if (skip == 0)
	{
	  if (fabs(end - start) / unit > 30.0)
	    skipfactor = 10;
	  else if (fabs(end - start) / unit >= 11.0)
	    skipfactor = 5;
	  else
	    skipfactor = 1;
	}
      else
	  skipfactor = skip;

      // crosspoint
      if (autocrosspoint)
	cp = start;
      else
	cp = crosspoint;

      for (int i = 0; i <= itercount; i++)
	{
	  fa = virstart + (double)i * unit;
	  if ((fa >= startval) && (fa <= endval))
	    {
	      labelcountfromzero = (int)(fabs(floor(fa / unit)) + .4);
	      if (labelcountfromzero % skipfactor == 0)
		{
		  NormalizeTickLabel(t->text, &normfactor, start, end, unit, fa,
				     true, format, addplus
				 ,autoprecision, intprecision, decimalprecision);
		  
		  w = (int)((double)t->gettextwidth());
		  ir.left    = 0;
		  ir.right   = w;
		  ir.top     = -h;
		  ir.bottom  = t->fo.size - h;
		  BBboxrotatedrect(&ir, fo.angle);
		  
		  if (cp - cprange * unit < fa &&
		      fa < cp + cprange * unit)
		    {
		      switch(crosspointparity)
			{
			case CROSSPOINTPARITY_CENTER:
			  lateralposition = (int)virtual_to_axis(fa) - (ir.left + ir.right) / 2;
			  break;
			case CROSSPOINTPARITY_MINUS:
			  lateralposition = (int)virtual_to_axis(fa) - ir.right
			    - crosspointoffset;
			  break;
			case CROSSPOINTPARITY_PLUS:
			  lateralposition = (int)virtual_to_axis(fa) - ir.left
			    + crosspointoffset;
			  break;
			}
		    }
		  else
		    lateralposition = (int)virtual_to_axis(fa) - (ir.left + ir.right) / 2;
		  
		  if (labelparity == 1)
		    tempdpoint = axis_to_world(lateralposition,
					       tickend - ir.top + offset);
		  else
		    tempdpoint = axis_to_world(lateralposition,
					       tickstart - ir.bottom - offset);
		  
		  t->p1.x = (int)tempdpoint.x;
		  t->p1.y = (int)tempdpoint.y;
		  
		  if (cp - cprange * unit < fa &&  fa < cp + cprange * unit)
		    {
		      if(crosspointparity != CROSSPOINTPARITY_NULL)
			t->drawtext(flag, f, 0, 0);
		    }
		  else
		    t->drawtext(flag, f, 0, 0);
		}
	    }
	}
      break;
    case SCALE_LOG:
    case SCALE_INV:
      virstart = pow(10.0, floor(log10(start)));
      virend   = pow(10.0, floor(log10(end)));
      virend += 1.0;
      delta           = DELTA_AXIS * (log10(end) - log10(start));
      startval        = start      / pow(10.0, delta);
      endval          = end        * pow(10.0, delta);
      
      itercount = (int)(log10(virend / virstart)) + 2;

      // set skip factor
      if (skip == 0)
	{
	  if (log10(end) - log10(start) >= 60.0)
	    skipfactor = 10;
	  else if (log10(end) - log10(start) >= 18.0)
	    skipfactor = 5;
	  else if (log10(end) - log10(start) >= 12.0)
	    skipfactor = 3;
	  else if (log10(end) - log10(start) >= 7.0)
	    skipfactor = 2;
	  else
	    skipfactor = 1;
	}
      else
	  skipfactor = skip;
      
      for (int i = 0; i <= itercount; i++)
	{
	  fa = virstart * pow(10.0, (double)i);
	  if ((fa >= startval) && (fa <= endval))
	    {
	      labelcountfromzero = (int)floor((log10(fa)+.4));
	      if (labelcountfromzero % skipfactor == 0)
		{
		  NormalizeTickLabel(t->text, &normfactor, start, end, 10, fa,
				     false, format, addplus
				     ,autoprecision, intprecision, decimalprecision);
		  
		  w = (int)((double)t->gettextwidth());
		  ir.left    = 0;
		  ir.right   = w;
		  
		  if (format == LABEL_FORMAT_EXP)
		    {
		      ir.top     = -(int)((double)h * 1.2);
		      ir.bottom  = t->fo.size - (int)((double)h * 1.2);
		    }
		  else
		    {
		      ir.top     = -h;
		      ir.bottom  = t->fo.size - h;
		    }
		  BBboxrotatedrect(&ir, fo.angle);
		  
		  if (cp - cprange * unit < fa &&
		      fa < cp + cprange * unit)
		    {
		      switch(crosspointparity)
			{
			case CROSSPOINTPARITY_CENTER:
			  lateralposition = (int)virtual_to_axis(fa) - (ir.left + ir.right) / 2;
			  break;
			case CROSSPOINTPARITY_MINUS:
			  lateralposition = (int)virtual_to_axis(fa) - ir.right
			    - crosspointoffset;
			  break;
			case CROSSPOINTPARITY_PLUS:
			  lateralposition = (int)virtual_to_axis(fa) - ir.left
			    + crosspointoffset;
			  break;
			}
		    }
		  else
		    lateralposition = (int)virtual_to_axis(fa) - (ir.left + ir.right) / 2;

		  if (labelparity == 1)
		    tempdpoint = axis_to_world(lateralposition,
					       tickend - ir.top + offset);
		  else
		    tempdpoint = axis_to_world(lateralposition,
					       tickstart - ir.bottom - offset);
		  
		  t->p1.x = (int)tempdpoint.x;
		  t->p1.y = (int)tempdpoint.y;

		  if (cp - cprange * unit < fa &&  fa < cp + cprange * unit)
		    {
		      if(crosspointparity != CROSSPOINTPARITY_NULL)
			t->drawtext(flag, f, 0, 0);
		    }
		  else
		    t->drawtext(flag, f, 0, 0);
		}
	    }
	}
      break;
    }

  // normalize factor
  if (normfactor != std::string(""))
    {
      *t->text = normfactor;
      w = (int)((double)t->gettextwidth());
      ir.left    = 0;
      ir.right   = w;
      ir.top     = -h;
      ir.bottom  = t->fo.size - h;

      BBboxrotatedrect(&ir, fo.angle);
      
      if (labelparity == 1)
	tempdpoint = axis_to_world(length - (ir.right - ir.left),
				   tickend - ir.top + t->fo.size + offset);
      else
	tempdpoint = axis_to_world(length - (ir.right - ir.left),
				   tickstart - ir.bottom - t->fo.size - offset);

      
      t->p1.x = (int)tempdpoint.x;
      t->p1.y = (int)tempdpoint.y;
      t->drawtext(flag, f, 0, 0);
    }
  
  
  delete t;
  return true;
  
}

bool axis::readaxis(FILE *fp)
{
  char stemp[10000];
  std::string str;
  char *p;
  int readversion;
  

  // version
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (nextitem(stemp, &str) != 0)
      readversion = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // id
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (nextitem(stemp, &str) != 0)
      {
	if (readversion >= 4)
	  id = atoi(str.c_str());
      }
    else
      return false;
  }
  else
    return false;

  // axistype, display, commonset, scaling, invertscale
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // axistype
    p = nextitem(p, &str);
    if (p != 0)
      axistype = atoi(str.c_str());
    else
      return false;
    // display
    p = nextitem(p, &str);
    if (p != 0)
      display = atoi(str.c_str());
    else
      return false;
    // commonset
    p = nextitem(p, &str);
    if (p != 0)
      commonset = atoi(str.c_str());
    else
      return false;
    // scaling
    p = nextitem(p, &str);
    if (p != 0)
      scaling = atoi(str.c_str());
    else
      return false;
    // invertscale
    p = nextitem(p, &str);
    if (p != 0)
      invertscale = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;


  // origin.x, origin.y, length, angl
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // origin.x
    p = nextitem(p, &str);
    if (p != 0)
      origin.x = atoi(str.c_str());
    else
      return false;
    // origin.y
    p = nextitem(p, &str);
    if (p != 0)
      origin.y = atoi(str.c_str());
    else
      return false;
    // length
    p = nextitem(p, &str);
    if (p != 0)
      length = atoi(str.c_str());
    else
      return false;
    // angle
    p = nextitem(p, &str);
    if (p != 0)
      angle = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // basecolor
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // basecolor.red
    p = nextitem(p, &str);
    if (p != 0)
      basecolor.red = atoi(str.c_str());
    else
      return false;
    // basecolor.green
    p = nextitem(p, &str);
    if (p != 0)
      basecolor.green = atoi(str.c_str());
    else
      return false;
    // basecolor.blue
    p = nextitem(p, &str);
    if (p != 0)
      basecolor.blue = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // baselinestyle
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // framelinestyle.width
    p = nextitem(p, &str);
    if (p != 0)
      baselinestyle.width = atoi(str.c_str());
    else
      return false;
    // baselinestyle.cap
    p = nextitem(p, &str);
    if (p != 0)
      baselinestyle.cap = atoi(str.c_str());
    else
      return false;
    // baselinestyle.join
    p = nextitem(p, &str);
    if (p != 0)
      baselinestyle.join = atoi(str.c_str());
    else
      return false;
    // baselinestyle.style
    p = nextitem(p, &str);
    if (p != 0)
      baselinestyle.style = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // tickarea
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // tickarea_start
    p = nextitem(p, &str);
    if (p != 0)
      tickarea_start = atof(str.c_str());
    else
      return false;
    // tickarea_end
    p = nextitem(p, &str);
    if (p != 0)
      tickarea_end = atof(str.c_str());
    else
      return false;
  }
  else
    return false;

  // start, end, unit, crosspoint, autocrosspoint
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // start
    p = nextitem(p, &str);
    if (p != 0)
      start = atof(str.c_str());
    else
      return false;
    // end
    p = nextitem(p, &str);
    if (p != 0)
      end = atof(str.c_str());
    else
      return false;
    // unit
    p = nextitem(p, &str);
    if (p != 0)
      unit = atof(str.c_str());
    else
      return false;
    // crosspoint
    p = nextitem(p, &str);
    if (p != 0)
      crosspoint = atof(str.c_str());
    else
      return false;
    // autocrosspoint
    p = nextitem(p, &str);
    if (p != 0)
      autocrosspoint = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // tickparity, tickwidth, ticklength, divisio
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // tickparity
    p = nextitem(p, &str);
    if (p != 0)
      tickparity = atoi(str.c_str());
    else
      return false;
    // tickwidth
    p = nextitem(p, &str);
    if (p != 0)
      tickwidth = atoi(str.c_str());
    else
      return false;
    // ticklength
    p = nextitem(p, &str);
    if (p != 0)
      ticklength = atoi(str.c_str());
    else
      return false;
    // division
    p = nextitem(p, &str);
    if (p != 0)
      division = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // majorticklinecolor
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // majorticklinecolor.red
    p = nextitem(p, &str);
    if (p != 0)
      majorticklinecolor.red = atoi(str.c_str());
    else
      return false;
    // majorticklinecolor.green
    p = nextitem(p, &str);
    if (p != 0)
      majorticklinecolor.green = atoi(str.c_str());
    else
      return false;
    // majorticklinecolor.blue
    p = nextitem(p, &str);
    if (p != 0)
      majorticklinecolor.blue = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // minorticklinecolor
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // minorticklinecolor.red
    p = nextitem(p, &str);
    if (p != 0)
      minorticklinecolor.red = atoi(str.c_str());
    else
      return false;
    // minorticklinecolor.green
    p = nextitem(p, &str);
    if (p != 0)
      minorticklinecolor.green = atoi(str.c_str());
    else
      return false;
    // minorticklinecolor.blue
    p = nextitem(p, &str);
    if (p != 0)
      minorticklinecolor.blue = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // majorticklinestyle
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // width
    p = nextitem(p, &str);
    if (p != 0)
      majorticklinestyle.width = atoi(str.c_str());
    else
      return false;
    // cap
    p = nextitem(p, &str);
    if (p != 0)
      majorticklinestyle.cap = atoi(str.c_str());
    else
      return false;
    // join
    p = nextitem(p, &str);
    if (p != 0)
      majorticklinestyle.join = atoi(str.c_str());
    else
      return false;
    // style
    p = nextitem(p, &str);
    if (p != 0)
      majorticklinestyle.style = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // minorticklinestyle
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // width
    p = nextitem(p, &str);
    if (p != 0)
      minorticklinestyle.width = atoi(str.c_str());
    else
      return false;
    // cap
    p = nextitem(p, &str);
    if (p != 0)
      minorticklinestyle.cap = atoi(str.c_str());
    else
      return false;
    // join
    p = nextitem(p, &str);
    if (p != 0)
      minorticklinestyle.join = atoi(str.c_str());
    else
      return false;
    // style
    p = nextitem(p, &str);
    if (p != 0)
      minorticklinestyle.style = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;


  // label
  // fo
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // face
    p = nextitem(p, &str);
    if (p != 0)
      fo.face = atoi(str.c_str());
    else
      return false;

    if (readversion >= 2)
      {
	// jpface
	p = nextitem(p, &str);
	if (p != 0)
	  fo.jpface = atoi(str.c_str());
	else
	  return false;
      }

    // size
    p = nextitem(p, &str);
    if (p != 0)
      fo.size = atoi(str.c_str());
    else
      return false;
    // angle
    p = nextitem(p, &str);
    if (p != 0)
      fo.angle = atof(str.c_str());
    else
      return false;
  }
  else
    return false;

  // fontcolor
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // red
    p = nextitem(p, &str);
    if (p != 0)
      fontcolor.red = atoi(str.c_str());
    else
      return false;
    // green
    p = nextitem(p, &str);
    if (p != 0)
      fontcolor.green = atoi(str.c_str());
    else
      return false;
    // blue
    p = nextitem(p, &str);
    if (p != 0)
      fontcolor.blue = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // format
    p = nextitem(p, &str);
    if (p != 0)
      format = atoi(str.c_str());
    else
      return false;
    // addplus
    p = nextitem(p, &str);
    if (p != 0)
      addplus = atoi(str.c_str());
    else
      return false;
    // offset
    p = nextitem(p, &str);
    if (p != 0)
      offset = atoi(str.c_str());
    else
      return false;
    // skip
    p = nextitem(p, &str);
    if (p != 0)
      skip = atoi(str.c_str());
    else
      return false;
    // autoprecision
    p = nextitem(p, &str);
    if (p != 0)
      autoprecision = atoi(str.c_str());
    else
      return false;
    // intprecision
    p = nextitem(p, &str);
    if (p != 0)
      intprecision = atoi(str.c_str());
    else
      return false;
    // decimalprecision
    p = nextitem(p, &str);
    if (p != 0)
      decimalprecision = atoi(str.c_str());
    else
      return false;
    // labelparity
    p = nextitem(p, &str);
    if (p != 0)
      labelparity = atoi(str.c_str());
    else
      return false;
    // crosspointparity
    if (readversion >= 4)
      {
	p = nextitem(p, &str);
	if (p != 0)
	  crosspointparity = atoi(str.c_str());
	else
	  return false;
      }
    else
      crosspointparity = CROSSPOINTPARITY_CENTER; 
    // crosspointoffset
    p = nextitem(p, &str);
    if (p != 0)
      crosspointoffset = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  return true;
}


bool axis::writeaxis(FILE *fp)
{
  fprintf(fp, "[axis]\n");

  // base
  fprintf(fp, "%d\n", version);
  fprintf(fp, "%d\n", id);
  fprintf(fp, "%d %d %d %d %d\n", axistype, display, commonset, scaling, invertscale);
  fprintf(fp, "%d %d %d %f\n", origin.x, origin.y, length, angle);
  fprintf(fp, "%u %u %u\n", basecolor.red, basecolor.green, basecolor.blue);
  fprintf(fp, "%d %d %d %d\n", baselinestyle.width, baselinestyle.cap, baselinestyle.join, baselinestyle.style);
  fprintf(fp, "%.15g %.15g\n", tickarea_start, tickarea_end);

  // range
  fprintf(fp, "%.15g %.15g %.15g %.15g %d\n", start, end, unit, crosspoint, autocrosspoint);

  //tick, subtick
  fprintf(fp, "%d %d %d %d\n", tickparity, tickwidth, ticklength, division);
  fprintf(fp, "%u %u %u\n", majorticklinecolor.red, majorticklinecolor.green, majorticklinecolor.blue);
  fprintf(fp, "%u %u %u\n", minorticklinecolor.red, minorticklinecolor.green, minorticklinecolor.blue);
  fprintf(fp, "%d %d %d %d\n", majorticklinestyle.width, majorticklinestyle.cap, majorticklinestyle.join, majorticklinestyle.style);
  fprintf(fp, "%d %d %d %d\n", minorticklinestyle.width, minorticklinestyle.cap, minorticklinestyle.join, minorticklinestyle.style);

  //label
  fprintf(fp, "%d %d %d %.15g\n", fo.face, fo.jpface, fo.size, fo.angle);
  fprintf(fp, "%u %u %u\n", fontcolor.red, fontcolor.green, fontcolor.blue);
  fprintf(fp, "%d %d %d %d %d %d %d %d %d %d\n", format, addplus, offset, skip, autoprecision,
	  intprecision, decimalprecision, labelparity, crosspointparity, crosspointoffset);

  return true;
}








