/*----------------------------------------------------------------------------
                      linearfit.cc (linear fitting routine)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string>

#include "linearfit.h"
#include "naninf.h"
#include "calc.h"
#include "graph.h"
#include "frame.h"
#include "data.h"

extern int CalcFormulastring(char *s, double *n, int InitFormula, data* dat, int cellno, int seriesno, int Isminmax);

bool getlinearfitfunctionname(int id, std::string *out)
{
  switch(id)
    {
    case FIT_POLY00:
      *out = std::string("0-poly.");
      break;
    case FIT_POLY01:
      *out = std::string("1-poly.");
      break;
    case FIT_POLY02:
      *out = std::string("2-poly.");
      break;
    case FIT_POLY03:
      *out = std::string("3-poly.");
      break;
    case FIT_POLY04:
      *out = std::string("4-poly.");
      break;
    case FIT_POLY05:
      *out = std::string("5-poly.");
      break;
    case FIT_POLY06:
      *out = std::string("6-poly.");
      break;
    case FIT_POLY07:
      *out = std::string("7-poly.");
      break;
    case FIT_POLY08:
      *out = std::string("8-poly.");
      break;
    case FIT_POLY09:
      *out = std::string("9-poly.");
      break;
    case FIT_POLY10:
      *out = std::string("10-poly.");
      break;
    case FIT_EXP:
      *out = std::string("A*exp(B*x)");
      break;
    case FIT_LOG:
      *out = std::string("A*ln(x) + B");
      break;
    case FIT_POW1:
      *out = std::string("A*x^B");
      break;
    case FIT_POW2:
      *out = std::string("A*B^x");
      break;
    default:
      *out = std::string("");
      return false;
    }
  return true;
}

bool CalcLinearLeastSq(doublearray *inx, 
		       doublearray *iny,
		       int n,                 // number of input data point
		       int m,	              // polynomial order
		       int mode,	      // fitting function type
		       double *factor,        // fitting factor
		       double *cor,           // correlation coefficient
		       double x0,             // fixed point
		       double y0,
		       bool Fixed,            // go through given fixed point or not
		       int StartLine,	      // start line for evaluation
		       int EndLine	      // end line for evaluation
		       )
{
  doublearray *r[12], *a[12], *inxrow = 0, *inyrow = 0;
  int i, j, k;
  int normalcount;
  double dbtemp = 0.0, sum, pivot, p, t;
  bool ErrorFlag = false;
  double rss=0.0,ave=0.0,var=0.0;
  int tempstartline,tempendline;
  
  for (i = 0; i < 12; i++)
    {
      r[i] = 0;
      a[i] = 0;
    }
  
  for (i = 0; i < 12; i++)
    {
      r[i] = new doublearray;
      if (r[i] == 0)
	{
	  r[i] = 0;
	  ErrorFlag = true;
	  goto j1;
	}
    }
      
  for (i = 0; i < 12; i++)
    {
      a[i] = new doublearray;
      if (a[i] == 0)
	{
	  a[i] = 0;
	  ErrorFlag = true;
	  goto j1;
	}
    }
  
  inxrow = new doublearray;
  if (inxrow == 0)
    {
      inxrow = 0;
      ErrorFlag = true;
      goto j1;
    }

  inyrow = new doublearray;
  if (inyrow == 0)
    {
      inyrow = 0;
      ErrorFlag = true;
      goto j1;
    }
  
  // remove invalid data
  tempstartline = StartLine;
  if (EndLine == -1)
    tempendline = n;
  else
    {
      if (EndLine <= n)
	tempendline = EndLine;
      else
	tempendline = n;
    }

  normalcount = 0;
  inxrow->add(0);
  inyrow->add(0);
  for (i = tempstartline; i <= tempendline; i++)
    {
      if (isnormalvalue((*inx)[i - 1]) && isnormalvalue((*iny)[i - 1]))
	{
	  normalcount++;
	  inxrow->addat(normalcount, (*inx)[i - 1]);
	  inyrow->addat(normalcount, (*iny)[i - 1]);
	}
    }
  
  if (normalcount == 0)
    {
      ErrorFlag = true;
      goto j1;
    }
  
  n = normalcount;
  
  for (i = 0; i < 12; i++)
    r[i]->setarray(n + 1);

  // generate input function
  // in case of polynomial
  if (mode == 1)
    {
      if (!Fixed)
	{
	  for (i = 1; i <= n; i++)
	    (*r[0])[i] = 1.0;
	  
	  for (j = 1;j <= m; j++)
	    {
	      for (i = 1; i <= n; i++)
		{
		  dbtemp = (*r[j - 1])[i] * (*inxrow)[i];

		  if (isnormalvalue(dbtemp))
		    {
		      (*r[j])[i] = dbtemp;
		    }
		  else
		    {
		      ErrorFlag = true;
		      goto j1;
		    }
		}
	    }
	}
      else if (Fixed)
	{
	  if (m == 0)
	    {
	      ErrorFlag = true;
	      goto j1;
	    }
	  
	  m--;
	  for (j = 0; j <= m; j++)
	    {
	      for (i = 1; i <= n; i++)
		{
		  dbtemp = pow((*inxrow)[i], j +1 ) - pow(x0, j + 1);
		  if (isnormalvalue(dbtemp))
		    {
		      (*r[j])[i] = dbtemp;
		    }
		  else
		    {
		      ErrorFlag = true;
		      goto j1;
		    }
		}
	    }
	}
    }
  
  //in case of exponential (y=Aexp(Bx) -> lny=lnA+Bx)
  else if (mode == 2)
    {
      if (!Fixed)
	{
	  m = 1;
	  for (i = 1; i <= n; i++)
	    {
	      (*r[0])[i] = 1.0;
	      (*r[1])[i] = (*inxrow)[i];
	    }
	}
      else if (Fixed)
	{
	  m = 0;
	  for (i = 1; i <= n; i++)
	    (*r[0])[i] = (*inxrow)[i] - x0;
	}
    }
  
  //in case of logalithm (y=A + Bln x )
  else if (mode == 3)
    {
      if (!Fixed)
	{
	  m = 1;
	  for (i = 1;i <= n; i++)
	    {
	      if ((*inxrow)[i] <= 0.0)
		{
		  ErrorFlag = true;
		  goto j1;
		}
	      
	      (*r[0])[i] = 1.0;
	      (*r[1])[i] = log((*inxrow)[i]);
	    }
	}
      else if (Fixed)
	{
	  if (x0 <= 0.0)
	    {
	      ErrorFlag = true;
	      goto j1;
	    }
	  
	  m = 0;
	  for (i = 1; i <= n; i++)
	    {
	      if ((*inxrow)[i] <= 0.0)
		{
		  ErrorFlag = true;
		  goto j1;
		}
	      
	      (*r[0])[i] = log((*inxrow)[i]) - log(x0);
	    }
	}
    }
  // in case of power1 (y = A*b^x -> ln y = ln A + ln B * x )
  else if (mode == 4)
    {
      if (!Fixed)
	{
	  m = 1;
	  for (i = 1; i <= n; i++)
	    {
	      (*r[0])[i] = 1.0;
	      (*r[1])[i] = (*inxrow)[i];
	    }
	}
      else if (Fixed)
	{
	  m = 0;
	  for (i = 1; i <= n; i++)
	    (*r[0])[i] = (*inxrow)[i] - x0;
	}
    }
  //in case of power2 (y = A*x^B -> ln y = ln A + B * ln x )
  else if (mode == 5)
    {
      if (!Fixed)
	{
	  m = 1;
	  for (i = 1; i <= n; i++)
	    {
	      (*r[0])[i] = 1.0;
	      (*r[1])[i] = log((*inxrow)[i]);
	    }
	}
      else if (Fixed)
	{
	  if (x0 <= 0.0)
	    {
	      ErrorFlag = true;
	      goto j1;
	    }
	  
	  m = 0;
	  for (i = 1; i <= n; i++)
	    {
	      if ((*inxrow)[i] <= 0.0)
		{
		  ErrorFlag = true;
		  goto j1;
		}
	      
	      (*r[0])[i] = log((*inxrow)[i]) - log(x0);
	    }
	}
    }
  
  for (i = 0; i < 12; i++)
    a[i]->setarray(n);

  for (i = 0; i <= n; i++)
    {
      for (j = i; j <= m; j++)
	{
	  sum = 0.0;
	  for (k = 1; k <= n; k++)
	    {
	      dbtemp = sum + (*r[i])[k] * (*r[j])[k];
	      if (isnormalvalue(dbtemp))
		{
		  sum = dbtemp;
		}
	      else
		{
		  ErrorFlag = true;
		  goto j1;
		}
	    }
	  
	  (*a[j])[i] = sum;
	  (*a[i])[j] = sum;
	}
    }
  
  for (i = 0; i <= m; i++)
    {
      sum = 0.0;
      for (k = 1; k <= n; k++)
	{
	  if (!Fixed)
	    {
	      if (mode == 1 || mode == 3)
		dbtemp = sum + (*r[i])[k] * (*inyrow)[k];
	      else if (mode == 2 || mode == 4 || mode == 5)
		{
		  if ((*inyrow)[k] <= 0.0)
		    {
		      ErrorFlag = true;
		      goto j1;
		    }
		  dbtemp = sum + (*r[i])[k] * log((*inyrow)[k]);
		}
	    }
	  else
	    {
	      if (mode == 1 || mode == 3)
		dbtemp = sum + (*r[i])[k] * ((*inyrow)[k] - y0);
	      else if (mode == 2 || mode == 4 || mode == 5)
		{
		  if ((*inyrow)[k] <= 0.0 || y0 <= 0)
		    {
		      ErrorFlag = true;
		      goto j1;
		    }
		  dbtemp = sum + (*r[i])[k] * (log((*inyrow)[k]) - log(y0));
		}
	    }
	  
	  if (isnormalvalue(dbtemp))
	    {
	      sum = dbtemp;
	    }
	  else
	    {
	      ErrorFlag = true;
	      goto j1;
	    }
	  
	}
      
      (*a[m+1])[i] = sum;
    }
  
  for (k = 0; k <= m; k++)
    {
      pivot = (*a[k])[k];
      
      if (pivot == 0.0)
	{
	  ErrorFlag = true;
	  goto j1;
	}
      
      for (j = k; j <= m + 1; j++)
	(*a[j])[k] /= pivot;
      
      for (i = 0; i <= m; i++)
	{
	  if (i != k)
	    {
	      t = (*a[k])[i];
	      for (j = k; j <= m + 1; j++)
		{
		  dbtemp = (*a[j])[i] - t * (*a[j])[k];
		  if (isnormalvalue(dbtemp))
		    {
		      (*a[j])[i] = dbtemp;
		    }
		  else
		    {
		      ErrorFlag = true;
		      goto j1;
		    }
		}
	    }
	}
    }
  
  // calculate coefficients
  for (i = 0; i <= m; i++)
    factor[i] = (*a[m + 1])[i];
  
  // polynomial
  if (mode == 1)
    {
      if (Fixed)
	{
	  m++;
	  for (i = m; i >= 1; i--)
	    factor[i] = factor[i - 1];
	  
	  factor[0] = y0;
	  for (i = 1; i <= m; i++)
	    factor[0] -= factor[i] * pow(x0, i);
	}
    }
  else if (mode == 2)
    {
      m = 1;
      if (!Fixed)
	factor[0] = exp(factor[0]);
      else
	{
	  factor[1] = factor[0];
	  factor[0] = log(y0) - factor[1] * x0; factor[0] = exp(factor[0]);
	}
      
    }
  else if (mode == 3)
    {
      m = 1;
      if (Fixed)
	{
	  if (x0 <= 0.0)
	    {
	      ErrorFlag = true;
	      goto j1;
	    }
	  
	  factor[1] = factor[0];
	  factor[0] = y0 - factor[1] * log(x0);
	}
    }
  else if (mode == 4)
    {
      m = 1;
      if (!Fixed)
	{
	  factor[0] = exp(factor[0]);
	  factor[1] = exp(factor[1]);
	}
      else if (Fixed)
	{
	  factor[1] = exp(factor[0]);
	  if (factor[1] <= 0.0)
	    {
	      ErrorFlag = true;
	      goto j1;
	    }
	  factor[0] = exp(log(y0) - log(factor[1]) * x0);
	}
    }
  else if (mode == 5)
    {
      m = 1;
      if (!Fixed)
	factor[0] = exp(factor[0]);
      else if (Fixed)
	{
	  factor[1] = factor[0];
	  factor[0] = log(y0) - factor[1] * log(x0); factor[0] = exp(factor[0]);
	}
    }

  // correlation coefficient
  if (!Fixed)
    {
      for (j = 0; j < n; j++)
	{
	  if (mode == 1)
	    {
	      p = factor[m];
	      for (i = m - 1; i >= 0; i--)
		p = p * (*inxrow)[j + 1] + factor[i];
	      rss += (p - (*inyrow)[j + 1]) * (p - (*inyrow)[j + 1]);
	      ave += (*inyrow)[j + 1];
	    }
	  else if (mode == 2)
	    {
	      p = (*inxrow)[j + 1] * factor[1];
	      p += log(factor[0]);
	      
	      rss += (p - log((*inyrow)[j + 1])) * (p - log((*inyrow)[j + 1]));
	      ave += log((*inyrow)[j + 1]);
	    }
	  else if (mode == 3)
	    {
	      p = log((*inxrow)[j + 1]);
	      p *= factor[1];
	      p += factor[0];
	      
	      rss += (p - (*inyrow)[j + 1]) * (p - (*inyrow)[j + 1]);
	      ave += (*inyrow)[j + 1];
	    }
	  else if (mode == 4)
	    {
	      p = (*inxrow)[j + 1] * log(factor[1]);
	      p += log(factor[0]);
	      
	      rss += (p - log((*inyrow)[j + 1])) * (p - log((*inyrow)[j + 1]));
	      ave += log((*inyrow)[j + 1]);
	    }
	  else if (mode == 5)
	    {
	      p = log((*inxrow)[j + 1]) * factor[1];
	      p += log(factor[0]);
	      
	      rss += (p - log((*inyrow)[j + 1])) * (p - log((*inyrow)[j + 1]));
	      ave += log((*inyrow)[j + 1]);
	    }
	}
      
      ave /= n;
      for (j = 0; j < n; j++)
	{
	  if (mode == 1 || mode == 3)
	    {
	      if (!Fixed)
		var += ((*inyrow)[j + 1] - ave) * ((*inyrow)[j + 1] - ave);
	      else
		var += ((*inyrow)[j + 1] - y0 - ave) * ((*inyrow)[j + 1] -y0 - ave);
	    }
	  else if(mode == 2 || mode == 4 || mode == 5)
	    {
	      if (!Fixed)
		var += (log((*inyrow)[j + 1]) - ave) * (log((*inyrow)[j + 1]) - ave);
	      else
		var += (log((*inyrow)[j + 1]) - log(y0) - ave) * (log((*inyrow)[j + 1]) - log(y0) - ave);
	    }
	}
      
      *cor = sqrt(1 - rss / var);
    }
 j1:;
  
  if (inxrow != 0)
    delete inxrow;
  
  if (inyrow != 0)
    delete inyrow;
  
  for (i = 0; i < 12; i++)
    {
      if (r[i] != 0)
	delete r[i];
    }
  
  for (i = 0; i < 12; i++)
    {
      if (a[i] != 0)
	delete a[i];
    }
  
  if (ErrorFlag)
    return false;
  else
    return true;
}

bool data::ExecLinearSQ(int fittingCurve, bool Fixed, double fixed_x, double fixed_y, int startline, int endline, std::string* outfnstr, std::string *infostr)
{
  doublearray *inx = 0, *iny = 0;
  int inputno, i;
  bool ErrorFlag = false;
  double factor[20];
  double cor;
  double x0;
  double y0;
  int StartLine;
  int EndLine;
  char stemp[256];
  std::string temp_pre_formula, temp_pre_formula_x, temp_pre_formula_y;
  std::string mes;
  int NumPoint;
  double fa, fb;
  frame *fr = (frame *)parent;
  std::string outfn;
  int m = 0;
  int mode = 1;

  inx = new doublearray;
  if (inx == 0)
    {
      inx = 0;
      ErrorFlag = true;
      goto j1;
    }

  iny = new doublearray;
  if (iny == 0)
    {
      iny = 0;
      ErrorFlag = true;
      goto j1;
    }


  //calclation
  NumPoint = x->getarraylength() - 1;
  if (trans_x)
    temp_pre_formula_x = *format_x;
  else
    temp_pre_formula_x = std::string("x");
  if (trans_y)
    temp_pre_formula_y = *format_y;
  else
    temp_pre_formula_y = std::string("y");
  
  for (i = 0; i <= NumPoint; i++)
    {
      CalcFormulastring((char *)temp_pre_formula_x.c_str(), &fa, true ,this, 0, fr->getdataindex(id), 0);
      CalcFormulastring((char *)temp_pre_formula_x.c_str(), &fa, false ,this, i, fr->getdataindex(id), 0);
      CalcFormulastring((char *)temp_pre_formula_y.c_str(), &fb, true ,this, 0, fr->getdataindex(id), 0);
      CalcFormulastring((char *)temp_pre_formula_y.c_str(), &fb, false ,this, i, fr->getdataindex(id), 0);
      if (isnormalvalue(fa) && (*x)[i].status == 0 &&
	  isnormalvalue(fb) && (*y)[i].status == 0
	  )
	{
	  if (!inx->add(fa))
	    goto j1;
	  if (!iny->add(fb))
	    goto j1;
	}
    }

  // data no.
  inputno = inx->getarraylength();

  // order
  switch (fittingCurve)
    {
    case 0:
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
    case 8:
    case 9:
    case 10:
      m = fittingCurve;
      break;
    case 11:
    case 12:
    case 13:
    case 14:
      m = 1;
      break;
    }
  //mode
  switch (fittingCurve)
    {
    case 0:
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
    case 8:
    case 9:
    case 10:
      mode = 1;
      break;
    case 11:
      mode = 2;
      break;
    case 12:
      mode = 3;
      break;
    case 13:
      mode = 4;
      break;
    case 14:
      mode = 5;
      break;
    }

  //fixed point X
  x0 = fixed_x;
  //fixed point Y
  y0 = fixed_y;

  //startline
  StartLine = startline;
  //endline
  EndLine = endline;

  if (CalcLinearLeastSq(inx
			,iny
			,inputno
			, m
			, mode
			, factor
			, &cor
			, x0
			, y0
			, Fixed
			, StartLine
			, EndLine
			))
	{

	  // generate message
	  // source data
	  mes = std::string("--------- fitting results ---------\\n");
	  mes += std::string("\\n");
	  mes += std::string("Original data: ");
	  mes += std::string(filename);
	  mes += std::string("\\n");
	  mes += std::string("\\n");
	  outfn = std::string("");
	  
	  switch(mode)
	    {
	    case 1:
	      mes += std::string("y = sigma Ai * x^i\\n");
	      mes += std::string("\\n");
	      for (i = 0; i <= m; i++)
		{
		  sprintf(stemp, "   A%d = %.15g\\n", i, factor[i]);
		  mes += std::string(stemp);
		}
	      for (i = m; i >= 0; i--)
		{
		  sprintf(stemp,"%+.15g*x^(%d)", factor[i], i);
		  outfn += std::string(stemp);
		}
	      break;
	    case 2:
	      mes += std::string("y = A * exp( B * x)\\n");
	      mes += std::string("\\n");
	      sprintf(stemp, "   A = %.15g\\n", factor[0]);
	      mes += std::string(stemp);
	      sprintf(stemp, "   B = %.15g\\n", factor[1]);
	      mes += std::string(stemp);
	      sprintf(stemp,"%+.15g*exp(%.15g*x)", factor[0], factor[1]);
	      outfn += std::string(stemp);
	      break;
	    case 3:
	      mes += std::string("y = A + B * ln( x )\\n");
	      mes += std::string("\\n");
	      sprintf(stemp, "   A = %.15g\\n", factor[0]);
	      mes += std::string(stemp);
	      sprintf(stemp, "   B = %.15g\\n", factor[1]);
	      mes += std::string(stemp);
	      sprintf(stemp,"%+.15g%+.15g*ln(x)", factor[0], factor[1]);
	      outfn += std::string(stemp);
	      break;
	    case 4:
	      mes += std::string("y = A * x ^ B\\n");
	      mes += std::string("\\n");
	      sprintf(stemp, "   A = %.15g\\n", factor[0]);
	      mes += std::string(stemp);
	      sprintf(stemp, "   B = %.15g\\n", factor[1]);
	      mes += std::string(stemp);
	      sprintf(stemp,"%.15g*%15g^x", factor[0], factor[1]);
	      outfn += std::string(stemp);
	      break;
	    case 5:
	      mes += std::string("y = A * B ^ x\\n");
	      mes += std::string("\\n");
	      sprintf(stemp, "   A = %.15g\\n", factor[0]);
	      mes += std::string(stemp);
	      sprintf(stemp, "   B = %.15g\\n", factor[1]);
	      mes += std::string(stemp);
	      sprintf(stemp,"%.15g*x^(%.15g)", factor[0], factor[1]);
	      outfn += std::string(stemp);
	      break;
	    }
	  if (!Fixed)
	    {
	      mes += std::string("\\n");
	      sprintf(stemp, "   Corr. Coeff. = %.15g\\n", cor);
	      mes += std::string(stemp);
	    }
	  
	  mes += std::string("\\n");
	  mes += std::string("\\n");
	  
	  sprintf(stemp, "Fitting line: %d -> %d\\n", StartLine, EndLine);
	  mes += std::string(stemp);

	  if (Fixed)
	  {
	    mes += std::string("Go through fixed point: true\\n");
	    sprintf(stemp, "Fixed point: (%.15g, %.15g)\\n", x0, y0);
	    mes += std::string(stemp);
	  }
	  mes += std::string("\\n");
	  mes += std::string("--------------- end ---------------\\n");

	  *outfnstr = outfn;
	  *infostr = mes;
	}
  else
    {
      ErrorFlag = true;
      goto j1;
    }
 j1:;

  if (inx != 0)
    delete inx;
  if (iny != 0)
    delete iny;
  
  if (ErrorFlag)
    return false;
  else
    return true;
}

